local radius = constants.ball_r
local offset_factor = 4
local friction = 0.5
local spin_factor = 0.025
local init_speed = 2^9
local speed_increase = 2^5
local max_tangent = 1

local screen_w, screen_h = constants.screen_w, constants.screen_h
local paddle_w, paddle_l = constants.paddle_w, constants.paddle_l
local min_x, max_x = radius + paddle_w, screen_w - radius - paddle_w
local min_y, max_y = radius, screen_h - radius


Ball = class {
	x = 0,
	y = 0,
	a = 0,
	vx = 0,
	vy = 0,
	va = 0,
	counter = 0,
	waiting = true
}

local img1, img2, img_scale, img_offset

Ball[class.init] = function (self, game)
	self.game = game

	img1 = img1 or love.graphics.newImage("gfx/Ball.png")
	img2 = img2 or love.graphics.newImage("gfx/Ball-shading.png")
	img_scale = img_scale or 2*radius / img1:getWidth()
	img_offset = img_offset or img1:getWidth() / 2
end


local function bounce (self)
	self.counter = self.counter + 1
	self.speed = init_speed + self.counter * speed_increase
end

local function normalize_velocity (self)
	local vx, vy = self.vx, self.vy
	local ratio = self.speed / (vx^2 + vy^2) ^ 0.5
	self.vx, self.vy = vx * ratio, vy * ratio
end

local opposite = {
	l = 'r',
	r = 'l' 
}

local full_turn = 2 * math.pi

function Ball:update (dt)
	if self.waiting then
		return
	end
	
	-- Velocity phase
	-- Apply the Magnus effect (curve ball)
	local spin = self.va * spin_factor
	self.vy = self.vy + dt * self.vx * spin
	self.vx = self.vx - dt * self.vy * spin

	-- Position phase
	self.a = self.a + self.va * dt
	if self.a > full_turn or self.a < 0 then
		self.a = self.a % full_turn
	end

	self.x = self.x + self.vx * dt
	self.y = self.y + self.vy * dt

	-- Collision resolution
	local y = self.y
	local ext = (y < min_y and min_y) or (y > max_y and max_y)
	if ext then
		self.y = 2 * ext - y
		self.vy = -self.vy
	end

	local x, va, side = self.x
	if x < min_x then
		x, va, side = x - min_x, self.va, 'l'
	elseif x > max_x then
		x, va, side = max_x - x, -self.va, 'r'
	end

	if side then
		local paddle = self.game:getPaddle(side)
		local offset = self.y - paddle.y
		if offset >= -paddle_l/2 - radius and offset <= paddle_l/2 + radius then
			x = 0
			self.vx = -self.vx
			self.vy = self.vy + offset * offset_factor

			local target_va = (self.vy - paddle.vy) / radius
			va = va + (target_va - va) * friction

			bounce(self)

			local vel_tangent = self.vy / self.vx
			if vel_tangent > max_tangent or vel_tangent < -max_tangent then
				if vel_tangent > 0 then
					self.vy = max_tangent * self.vx
				else
					self.vy = -max_tangent * self.vx
				end
			end

			normalize_velocity(self)

			if side == 'l' then
				self.x, self.va = min_x + x, va
			else
				self.x, self.va = max_x - x, -va
			end
		elseif x < -paddle_w then
			self.game:getPaddle(opposite[side]).player:score()
			self:reset()
		end
	end
end

function Ball:draw ()
	local x, y = self.x, self.y
	love.graphics.setColor(255, 255, 255)
	love.graphics.draw(img1, x, y, self.a, img_scale, img_scale, img_offset, img_offset)
	love.graphics.draw(img2, x, y, 0, img_scale, img_scale, img_offset, img_offset)
end


function Ball:reset ()
	self.x, self.y = screen_w/2, screen_h/2
	self.vx, self.vy, self.va = 0, 0, 0
	self.counter = 0
	self.waiting = true
end

function Ball:launch (x, y)
	if self.waiting then
		self.vx, self.vy = x, y
		self.waiting = false
		self.speed = init_speed
		normalize_velocity(self)
	end
end


return Ball


-- vim: ts=4 sw=4 si
